(* Creates files which are included in the Makefile for the ocaml programs.
   Not part of the Browndye toolchain proper, but runs during the software 
   build.  Takes information from file .depend1, which is generated by
   ocamldepend, the ocaml dependency generator, and builds file .depend2,
   which is included in the Makefile.
*)


let end_chars ne str = 
  let n = String.length str in
    String.sub str (n-ne) ne
;;

let chopped_line ne str = 
  let n = String.length str in
    String.sub str 0 (n-ne)
;;

let input_lines chn = 
  let lines = ref [] in
    try
      while true do	
	let line = Lines.stripped (input_line chn) in
	  lines := line :: !lines
      done;
      List.rev !lines

    with End_of_file -> 
      List.rev !lines
;;

let lines = input_lines (open_in ".depend1");;

let pr = Printf.printf;;

let is_cmo line = 
  let tokens = Lines.split line in
    (end_chars 5 tokens.(0)) = ".cmo:"
;;

let straight_lines = 
  let rec loop lines cur_line final_lines = 
    match lines with
      | line :: rest ->
	  let capp line final_lines = 
	    if not (is_cmo line) then
	      line :: final_lines
	    else 
	      final_lines
	  in
	    if (end_chars 1 line) = "\\" then
	      let new_cur_line = cur_line ^ (chopped_line 1 line) in
		loop rest new_cur_line final_lines
	    else (
	      if cur_line = "" then
		loop rest "" (capp line final_lines)
	      else
		let new_cur_line = cur_line ^ line in
		  loop rest "" (capp new_cur_line final_lines)
	    )

      | [] -> final_lines
  in
  let rev_lines = loop lines "" [] in
    List.rev rev_lines
;;

let do_while f = 
  let rec loop () = 
    if f() then
      loop ()
  in
    loop()
;;



let dict = 
  let dict = Hashtbl.create 0 in
    List.iter
      (fun line ->
	let key, avalues = 
	  let tokens = Lines.split line in
	  let nt = Array.length tokens in 

	  let charray nc = 
	    Array.init (nt-nc) (fun i -> tokens.(i+nc)) 
	  in

	  let pkey = tokens.(0) in
	  if nt > 1 && tokens.(1) = ":" then
	    pkey, charray 2
	  else
	    chopped_line 1 pkey, charray 1
	      
	in
	if (end_chars 4 key) = ".cmx" then
	  let values = Array.to_list avalues 
	  in
	  (* trial *)
	  match values with
	    | [] -> ()
	    | _ -> 
	      Hashtbl.add dict key values;
      )
      straight_lines;

    let dependencies (key: string) =

      let rec deps key =  
	let values = 
	  if Hashtbl.mem dict key then
	    Hashtbl.find dict key 
	  else
	    []
	in
	  List.fold_left
	    (fun res value ->
	      res @ (value :: (deps value))
	    )
	    []
	    values
      in
      let raw = deps key in
	List.fold_left
	  (fun res value ->
	    if List.mem value res then
	      res
	    else
	      value :: res
	  )
	  []
	  (List.rev raw)
	  		  
    in
    let dict2 = Hashtbl.create 0 in
      Hashtbl.iter
	(fun key value ->
	  Hashtbl.add dict2 key (List.rev (dependencies key))
	)
	dict;
      dict2
	
;;

let out = open_out ".depend2";;

let fpr = Printf.fprintf;;

Hashtbl.iter
  (fun key values ->
    fpr out "%s: " key;
    List.iter 
      (fun value -> fpr out "%s " value)
      values;
    fpr out "\n";
  )
  dict
;;


let programs = 
  let input = Scanf.Scanning.from_file "program_list" in
  let lines = ref [] in
      do_while
	(fun () -> 
	  let line = Lines.stripped (Lines.readline input) in
	    if line = "" then
	      false
	    else (
	      lines := line :: !lines;
	      true
	    );
	);
    !lines
      
;;


let list_isect list0 list1 = 
  List.fold_left
    (fun res item0 ->
      res || (List.mem item0 list1)
    )
    false
    list0
;;

Hashtbl.iter
  (fun key values ->
    let pname = chopped_line 4 key in
      if List.mem pname programs then (
	fpr out "%s: " pname;
	List.iter 
	  (fun value -> 
	    if (end_chars 4 value) = ".cmx" then
	      fpr out "%s " value
	  )
	  values;
	fpr out "%s\n" key
      )      
  )
  dict
;;

fpr out "\n";;

let unix_progs = ["orchestrator.cmx"; "reactive_trajectories.cmx"];;
let expat_progs = ["jam_xml_ml_1f_parser.cmx";];;
let bigarray_progs = ["grid_parser.cmx"; "mpole_grid_fit.cmx"; "grid_distances.cmx"; "compute_effective_volumes.cmx"; "hydro_radius.cmx"; "lumped_charges.cmx"];;
let eigenvector_progs = ["eigenvectors.cmx"];;
let least_squares_progs = ["least_squares.cmx"];;

List.iter
  (fun program ->
    fpr out "%s:\n" program;
    fpr out "\tocamlopt -I ./ocaml-expat ";
    
    let cmx = program^".cmx" in
    let deps = cmx :: (Hashtbl.find dict cmx) in
      if list_isect deps unix_progs then
	fpr out "unix.cmxa ";
      if list_isect deps expat_progs then
	fpr out "expat.cmxa ";
      if list_isect deps bigarray_progs then
	fpr out "bigarray.cmxa ";
      if list_isect deps eigenvector_progs then
	fpr out "eigenvectors.cmxa ";
      if list_isect deps least_squares_progs then
	fpr out "least_squares.cmxa ";

    fpr out "$^ -o $@; rm -f ../bin/$@; ln -s ../aux/$@ ../bin/$@\n";
  )
  programs
;;
fpr out "\n";;


fpr out "PROGRAMS = ";;
List.iter
  (fun program ->
    fpr out "%s " program;
  )
  programs
;;
fpr out "\n";;

